/* MT_Addin.cpp - part 2 - MyopicTopics Addin
 * Written By: Jason Weiler
 * This code is in the public domain
 */
 
#define WIN32_LEAN_AND_MEAN

#include <windows.h>
#include <stdio.h>
#include <stddef.h>

#include "MT_Addin.h"

struct MyType
{
	int Integer1;
	int Integer2;
	char * BigString;
	char * StringTable[10];
	int StringIndex;
};

static HRESULT ReadStringFromDebuggee(DEBUGHELPER * pHelper, unsigned __int64 dlBaseAddress, DWORD dwMaxStringSize, char * szOutString, DWORD * pOutSizeRead);


/****************************/
/* HandleMyType_TwoIntegers */
/****************************/
ADDIN_API HRESULT WINAPI HandleMyType_TwoIntegers( DWORD /*dwAddress*/
									  , DEBUGHELPER *pHelper
									  , int nBase
									  , BOOL bUniStrings
									  , char *pResult
									  , size_t maxResult
									  , DWORD /*reserved*/ )
{
	// Step 1: Get the 64-bit pointer we want to inspect
	unsigned __int64 dlDataPtr = pHelper->GetRealAddress(pHelper);

	int Int1=0, Int2=0;
	DWORD dwNumRead=0;

	// Step 2: Read the first integer
	HRESULT ReadResult = pHelper->ReadDebuggeeMemoryEx(pHelper, dlDataPtr+offsetof(MyType,Integer1), 4, &Int1, &dwNumRead);

	// Step 3: Check for read Errors
	if (FAILED(ReadResult) || dwNumRead != 4)
	{
		strcpy_s(pResult, maxResult, "Error reading Integer1");
		return S_OK; // always return S_OK
	}

	// Step 4: Read the second integer
	ReadResult = pHelper->ReadDebuggeeMemoryEx(pHelper, dlDataPtr+offsetof(MyType,Integer2), 4, &Int2, &dwNumRead);

	// Step 5: Check for read Errors
	if (FAILED(ReadResult) || dwNumRead != 4)
	{
		strcpy_s(pResult, maxResult, "Error reading Integer2");
		return S_OK;
	}

	// Step 6: Compose out output string
	sprintf_s(pResult, maxResult, "%d + %d = %d", Int1, Int2, Int1+Int2);

	return S_OK;
}

/**************************/
/* HandleMyType_BigString */
/**************************/
ADDIN_API HRESULT WINAPI HandleMyType_BigString( DWORD /*dwAddress*/
									  , DEBUGHELPER *pHelper
									  , int nBase
									  , BOOL bUniStrings
									  , char *pResult
									  , size_t maxResult
									  , DWORD /*reserved*/ )
{
	// Get the 64-bit pointer we want to inspect
	unsigned __int64 dlDataPtr = pHelper->GetRealAddress(pHelper);

	DWORD dwNumRead=0;

	unsigned int BigStrPtr;
	HRESULT ReadResult = pHelper->ReadDebuggeeMemoryEx(pHelper, dlDataPtr+offsetof(MyType,BigString), 4, &BigStrPtr, &dwNumRead);
	if (FAILED(ReadResult) || dwNumRead != 4)
	{
		strcpy_s(pResult, maxResult, "Error reading BigString pointer");
		return S_OK;
	}

	char BigStr[1024];
	DWORD MaxToRead = min(1024, maxResult);

	// Read the StringIndex integer
	ReadResult = ReadStringFromDebuggee(pHelper, (unsigned __int64)BigStrPtr, MaxToRead, BigStr, &dwNumRead);
	if (FAILED(ReadResult))
	{
		strcpy_s(pResult, maxResult, "Error reading String Data");
		return S_OK;
	}

	sprintf_s(pResult, maxResult, "BigString == %s", BigStr);

	return S_OK;
}

/****************************/
/* HandleMyType_StringTable */
/****************************/
ADDIN_API HRESULT WINAPI HandleMyType_StringTable( DWORD /*dwAddress*/
									  , DEBUGHELPER *pHelper
									  , int nBase
									  , BOOL bUniStrings
									  , char *pResult
									  , size_t maxResult
									  , DWORD /*reserved*/ )
{
	// Step 1: Get the 64-bit pointer we want to inspect
	unsigned __int64 dlDataPtr = pHelper->GetRealAddress(pHelper);

	int iStrIdx=0;
	DWORD dwNumRead=0;

	// Step 2: Read the StringIndex integer
	HRESULT ReadResult = pHelper->ReadDebuggeeMemoryEx(pHelper, dlDataPtr+offsetof(MyType,StringIndex), 4, &iStrIdx, &dwNumRead);

	// Step 3: Check for read Errors
	if (FAILED(ReadResult) || dwNumRead != 4)
	{
		strcpy_s(pResult, maxResult, "Error reading StringIndex");
		return S_OK;
	}

	// Step 4: Bounds check the index value 
	if (iStrIdx < 0 || iStrIdx >= 10)
	{
		strcpy_s(pResult, maxResult, "StringIndex is out of bounds");
		return S_OK;
	}

	// Step 5: Calculate the offset into the string table at the index
	unsigned __int64 dlStrTblPtr = dlDataPtr + offsetof(MyType, StringTable) + (4*iStrIdx);

	// Step 6: Query the pointer value at that offset
	unsigned int StrPtr32;
	ReadResult = pHelper->ReadDebuggeeMemoryEx(pHelper, dlStrTblPtr, 4, &StrPtr32, &dwNumRead);

	// Step 7: Check for read Errors
	if (FAILED(ReadResult) || dwNumRead != 4)
	{
		strcpy_s(pResult, maxResult, "Error reading string table item pointer");
		return S_OK;
	}

	// Step 8: Handle a NULL pointer
	if (StrPtr32 == 0)
	{
		sprintf_s(pResult, maxResult, "StringTable[%d] == NULL", iStrIdx);
		return S_OK;
	}

	// Step 9: Query the string bytes at the pointer value
	char StrValue[64];
	DWORD MaxToRead = min(64, maxResult);

	ReadResult = ReadStringFromDebuggee(pHelper, (unsigned __int64)StrPtr32, MaxToRead, &StrValue[0], &dwNumRead);

	// Step 10: Check for read Errors
	if (FAILED(ReadResult))
	{
		strcpy_s(pResult, maxResult, "Error reading string item text");
		return S_OK;
	}

	// Step 11: Compose the output string and return
	sprintf_s(pResult, maxResult, "StringTable[%d] == %s", iStrIdx, StrValue);

	return S_OK;
}



HRESULT ReadStringFromDebuggee(DEBUGHELPER * pHelper, unsigned __int64 dlBaseAddress, DWORD dwMaxStringSize, char * szOutString, DWORD * pOutSizeRead)
{
	unsigned __int64 const dlAlignment = 1024;
	unsigned __int64 const dlPageMask = ~(dlAlignment-1);

	DWORD nGot;
	HRESULT rc = S_OK;
	*pOutSizeRead = 0;

	// Split our read up into parts if we cross a 1k boundary
	while ((dwMaxStringSize > 0) && (rc == S_OK))
	{
		unsigned __int64 dlEndAddr = dlBaseAddress+dwMaxStringSize;

		// If the data to read bridges one of our page boundaries (as specified by the dlAlignment argument)
		//   Then clip the end of the read to the end of the page. We'll loop back around to get more later.
		if ((dlBaseAddress & dlPageMask) != (dlEndAddr & dlPageMask))
		{ 
			dlEndAddr = (dlBaseAddress + dlAlignment) & dlPageMask;
		}

		DWORD dwReadSize = static_cast<DWORD>(dlEndAddr - dlBaseAddress);

		rc = pHelper->ReadDebuggeeMemoryEx(pHelper, dlBaseAddress, dwReadSize, szOutString, &nGot);

		// scan for the terminating null
		for(DWORD iChar=0; iChar < nGot; ++iChar)
		{
			if (szOutString[iChar] == '\0')
			{
				*pOutSizeRead += iChar;
				return rc;
			}
		}

		// total the characters we've read
		*pOutSizeRead += nGot;

		// set up for the next read
		dwMaxStringSize -= dwReadSize;
		szOutString = (char*)(reinterpret_cast<unsigned __int64>(szOutString) + dwReadSize);
		dlBaseAddress += dwReadSize;

	}

	// If we get here, then we've read a lot of data, but no terminating null.
	// We need to add one artificially
	if (*pOutSizeRead > 0)
	{
		*(szOutString-1) = '\0';
		--(*pOutSizeRead);
	}

	return rc;

}


/* Code from the project wizard (normally in dllmain.cpp) */
BOOL APIENTRY DllMain( HMODULE hModule,
                       DWORD  ul_reason_for_call,
                       LPVOID lpReserved
					 )
{
	switch (ul_reason_for_call)
	{
	case DLL_PROCESS_ATTACH:
	case DLL_THREAD_ATTACH:
	case DLL_THREAD_DETACH:
	case DLL_PROCESS_DETACH:
		break;
	}
	return TRUE;
}

